import re

import spacy
from spacy.tokens import Doc
from spacy.language import Language
from .tld import TLD


class Tokenizer:
    """Class for Turkish text tokenization. Applies a more robust url match pattern based
    on a list of valid top level domains.
    """

    def __init__(self, nlp: Language):
        self.nlp = spacy.blank('tr')
        self.nlp.max_length = nlp.max_length
        self.vocab = nlp.vocab

    def _is_url(self, text: str) -> bool:
        """Checks whether the string is a valid url or not.
        https://gist.github.com/gruber/8891611
        """
        tld = '|'.join(TLD)
        pattern = re.compile(
            fr'(?i)\b((?:https?:(?:/{{1,3}}|[a-z0-9%])|[a-z0-9.\-]+[.](?:{tld})/)'
            r'(?:[^\s()<>{}\[\]]+|\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\))+'
            r'(?:\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\)|[^\s`!()\[\]{{}};:\'"'
            fr'.,<>?«»“”‘’])|(?:(?<!@)[a-z0-9]+(?:[.\-][a-z0-9]+)*[.](?:{tld})\b/?(?!@)))'
        )
        return bool(pattern.match(text))

    def __call__(self, text):
        words, spaces = [], []
        pattern = '[({\[<)}\]>«»“”„‟‹›❝❞❟❠❮❯〝〞〟＂"‘’‚‛❛❜]'
        for token in self.nlp(text):
            # ensure whether the token is a valid url
            if token.like_url and not self._is_url(token.text):
                tokens = [t.text for t in self.nlp(token.text.replace('.', ' . '))
                          if not t.is_space]
                words.extend(tokens)
                spaces.extend([False] * len(tokens))
                spaces[-1] = bool(token.whitespace_)
            # fix bracket and quote tokenization errors, e.g. `bu"(bir)` -> `bu " ( bir )`
            elif re.search(fr'{pattern}', token.text):
                tokens = [t.text for t in self.nlp(re.sub(fr'({pattern})', r' \1 ', token.text))
                          if not t.is_space]
                words.extend(tokens)
                spaces.extend([False] * len(tokens))
                spaces[-1] = bool(token.whitespace_)
            else:
                words.append(token.text)
                spaces.append(bool(token.whitespace_))
        return Doc(self.vocab, words=words, spaces=spaces)



import os
import json
from pathlib import Path
from typing import Iterator, Union, Tuple, List, Dict, Any


class FS:
    """Class for file I/O and utilities.
    """
    @staticmethod
    def split_path(filepath: Union[str, Path]) -> Tuple[str, str, str]:
        """Extract file extension, file name and parent path.

        Args:
            filepath (Union[str, Path]): File path.

        Returns:
            Tuple[str, str, str]: Extracted extension, filename and parent path. 
        """
        parent, basename = os.path.split(str(filepath).rstrip(os.sep))
        filename, ext = os.path.splitext(basename)
        ext = ext[1:].strip().lower()
        return ext, filename, parent

    @staticmethod
    def read_json(filepath: Union[str, Path]) -> Union[Dict[str, Any], List[Any]]:
        """Read json file.

        Args:
            filepath (Union[str, Path]): File path.

        Returns:
            Union[Dict[str, Any], List[Any]]: File content.
        """
        with open(filepath, encoding='utf-8') as f:
            content = json.load(f)
        return content

    @staticmethod
    def write_json(
        content: Union[str, Dict[Any, Any], List[Any]],
        filepath: Union[str, Path],
        append: bool = False
    ) -> None:
        """Dump json file.

        Args:
            content (Union[Dict[Any, Any], List[Any]]): Content to be dumped.
            filepath (Union[str, Path]): File path.
            append (bool): Append to the existing file or not. Update key values 
                if file is a json dictionary and key exists. Defaults to False.
        """
        if not isinstance(content, (str, dict, list)):
            raise ValueError('Content must be a string, dictionary or list.')
        content = json.loads(content) if isinstance(content, str) else content
        if append and os.path.isfile(filepath):
            with open(filepath, encoding='utf-8') as f:
                data = json.load(f)
            if isinstance(data, dict):
                if not isinstance(content, dict):
                    raise ValueError('Content must be a dictionary.')
                data.update(content)
            elif isinstance(data, list):
                data.append(content)
            else:
                raise ValueError('An error occured while reading json file.')
            with open(filepath, 'w', encoding='utf-8') as f:
                json.dump(data, f, ensure_ascii=False)
        else:
            os.makedirs(os.path.dirname(filepath), exist_ok=True)
            with open(filepath, 'w', encoding='utf-8') as f:
                json.dump(content, f, ensure_ascii=False)

    @staticmethod
    def read_jsonl(filepath: Union[str, Path]) -> List[Union[Dict[str, Any], List[Any]]]:
        """Read jsonlines file.

        Args:
            filepath (Union[str, Path]): File path.

        Returns:
            Union[Dict[str, Any], List[Any]]: File content.
        """
        content = []
        with open(filepath, encoding='utf-8') as f:
            for line in f:
                content.append(json.loads(line.strip()))
        return content

    @staticmethod
    def write_jsonl(
        content: List[Union[Dict[Any, Any], List[Any]]],
        filepath: Union[str, Path],
        append: bool = False
    ) -> None:
        """Dump jsonlines file.

        Args:
            content (List[Union[Dict[Any, Any], List[Any]]]): Content to be dumped.
            filepath (Union[str, Path]): File path.
            append (bool): Append to the existing file or not. Defaults to False.
        """
        if not isinstance(content, list) or not all(isinstance(e, (dict, list)) for e in content):
            raise ValueError('Content must be a list of dictionaries or lists.')
        if append and os.path.isfile(filepath):
            with open(filepath, 'a+', encoding='utf-8') as f:
                for line in content:
                    json_record = json.dumps(line, ensure_ascii=False)
                    f.write(f'\n{json_record}')
        else:
            os.makedirs(os.path.dirname(filepath), exist_ok=True)
            with open(filepath, 'w', encoding='utf-8') as f:
                for line in content:
                    json_record = json.dumps(line, ensure_ascii=False)
                    f.write(f'{json_record}\n')
                f.truncate(f.tell() - len(os.linesep))

    @staticmethod
    def read_conll(filepath: Union[str, Path]) -> Iterator[List[Union[str, List[str]]]]:
        """Read conll file.

        Args:
            filepath (Union[str, Path]): File path.

        Yields:
            Iterator[List[Union[str, List[str]]]]: Yield conll file content,
                one for each sentence. 
        """
        with open(filepath, encoding="utf-8") as f:
            sents = f.read().strip().split('\n\n')
        for sent in sents:
            lines = []
            for line in [l.strip() for l in sent.strip().split('\n')]:
                lines.append(line if line.startswith('#') else line.split())
            yield lines

    @staticmethod
    def parse_conllu(
        filepath: Union[str, Path],
        min_tokens: int = 5
    ) -> List[Dict[str, List[str]]]:
        """Read conllu file, merge subtokens and normalize lemmas. 
        Lemma cases are inconsistent among treebanks.

        Args:
            filepath (Union[str, Path]): Conllu file path.
            min_tokens (int, optional): Minimum number of tokens the sentence should have.

        Returns:
            List[List[str]]: List of parsed sentences.
        """
        def lower(text: str) -> str:
            return text.replace('I', 'ı').replace('İ', 'i').lower()

        def isupper(text: str) -> str:
            return text == text.replace('i', 'İ').upper()

        def normalize_lemma(token, lemma):
            i, prefix = 1, ''
            while i <= len(lemma) and lower(token[:i]) == lower(lemma[:i]):
                prefix = token[:i]
                i += 1
            return prefix + lemma[len(prefix):]

        sents = []
        for lines in FS.read_conll(filepath):
            exclude = []
            for idx, line in enumerate(lines):
                if isinstance(line, str):
                    exclude.append(idx)
                    continue
                # merge subtokens
                if '-' in line[0]:
                    lines[idx+1][1] = line[1]
                    exclude.extend([idx, idx+2])
                # normalize lemma
                lines[idx][2] = normalize_lemma(line[1], line[2])
            lines = [l for idx, l in enumerate(lines) if idx not in exclude]
            if len(lines) >= min_tokens:
                # some sents are all uppercase, lower tokens and lemmas
                if all(isupper(l[1]) for l in lines):
                    for idx, line in enumerate(lines):
                        lines[idx][1] = lower(line[1])
                        lines[idx][2] = lower(line[2])
                sent = {k: [] for k in ('words', 'lemmas', 'poses', 'morphs')}
                for line in lines:
                    _, word, lemma, pos, _, morph, *_ = line
                    sent['words'].append(word)
                    sent['lemmas'].append(lemma)
                    sent['poses'].append(pos)
                    sent['morphs'].append(morph)
                sents.append(sent)
        return sents

    @staticmethod
    def read(filepath: Union[str, Path]) -> Union[str, Dict[str, Any], List[Any]]:
        """Read content from file.

        Args:
            filepath (Union[str, Path]): File path.

        Returns:
            Union[str, Dict[str, Any], List[Any]]: File content.
        """
        ext, *_ = FS.split_path(filepath)
        if not os.path.isfile(filepath):
            raise ValueError(f'File does not exist at path "{filepath}".')

        if ext.startswith('conll'):
            return FS.read_conll(filepath)
        elif ext == 'json':
            return FS.read_json(filepath)
        elif ext == 'jsonl':
            return FS.read_jsonl(filepath)
        else:
            with open(filepath, encoding='utf-8') as f:
                content = f.read()
            return content

    @staticmethod
    def to_disk(
        content: Union[str, Dict[Any, Any], List[Any]],
        filepath: Union[str, Path],
        append: bool = False
    ) -> None:
        """Write content to file.

        Args:
            content (Union[str, Dict[Any, Any], List[Any]]): File content to be written.
            filepath (Union[str, Path]): File path.
            append (bool, optional): Append to the existing file or not. Defaults to False.
        """
        ext, *_ = FS.split_path(filepath)
        if ext == 'json':
            FS.write_json(content, filepath, append=append)
        elif ext == 'jsonl':
            FS.write_jsonl(content, filepath, append=append)
        else:
            if append and os.path.isfile(filepath):
                with open(filepath, 'a+', encoding='utf-8') as f:
                    f.write(f'\n{content}')
            else:
                os.makedirs(os.path.dirname(filepath), exist_ok=True)
                with open(filepath, 'w', encoding='utf-8') as f:
                    f.write(content)



__version__ = '0.0.2'
__repo__ = 'https://github.com/nlpturk/nlpturk'
__model__ = 'nlpturk_model'
__download_url__ = f'{__repo__}/releases/download/v{__version__}/{__model__}-{__version__}-py3-none-any.whl'



import re
import sys
import itertools
import warnings
from pathlib import Path
from importlib.util import find_spec

import spacy
from spacy import util
from spacy.tokens.doc import Doc
from wasabi import Printer

from . import pkg
from .pipeline.tokenizer import Tokenizer
from .pipeline import sbd
from .doc import Document


class _M(sys.modules[__name__].__class__):
    """Class that makes the nlpturk package a callable module.
    """

    def __call__(self, text: str) -> Document:
        """Makes the nlpturk package callable.

        Usage: 
            import nlpturk
            doc = nlpturk(some_text)

            # iterate over tokens
            for token in doc:
                print(token.text, token.lemma, token.pos, token.is_sent_end)

            # or get tokens by token ids
            token = doc[i]

            # iterate over sentences
            for sent in doc.sents:
                print(sent.text, sent.vector)
                for token in sent:
                    print(token.text, token.lemma, token.pos, token.vector)

        Args:
            text (str): Text to be processed.

        Returns:
            Document: Document object.
        """
        if not hasattr(self, '_nlp'):
            warnings.filterwarnings('ignore')
            try:
                model_path = Path(find_spec(pkg.__model__).origin).parent
            except AttributeError:
                model_path = self._download()
            self._nlp = spacy.load(model_path)
            self._nlp.tokenizer = Tokenizer(self._nlp)

        return Document(self._get_doc(text))

    def _get_doc(self, text: str) -> Doc:
        """Removes whitespace tokens returned from Tokenizer and aligns character offsets
        of the tokens. 

        Args:
            text (str): Text to be processed.

        Returns:
            Doc: spaCy Doc object.
        """
        # grab trailing whitespaces for each word
        idx, ws = [0], []
        for m in re.finditer(r'\s+', text):
            if m.start(0) > 0:
                idx.append(m.end(0))
                ws.append(m.group(0))

        # grab leading whitespaces from text
        lws = ''.join(itertools.takewhile(str.isspace, text))
        idx[0] = len(lws) if lws else 0

        doc = self._nlp(' '.join(text.split()))

        # set aligned offsets and trailing whitespaces for each token
        tokens, i = [], idx[0]
        for t in doc:
            if t.whitespace_:
                tokens.append((i, ws[0]))
                idx.pop(0)
                ws.pop(0)
                i = idx[0]
            else:
                tokens.append((i, ''))
                i += len(t.text)

        # add trailing whitespaces to last token, if exist
        tokens[-1] = (tokens[-1][0], ws[0]) if ws else tokens[-1]

        doc.set_extension('idx', default=[t[0] for t in tokens], force=True)
        doc.set_extension('ws', default=[t[1] for t in tokens], force=True)
        doc.set_extension('lws', default=lws, force=True)

        return doc

    def _download(self) -> Path:
        """Downloads nlpTurk model.

        Args:
            model (str): nlpTurk model name.

        Returns:
            Path: Model path.
        """
        msg = Printer()
        msg.info(f'nlpTurk model not found! Downloading ...')
        cmd = [sys.executable, "-m", "pip", "install"] + [pkg.__download_url__]
        try:
            util.run_command(cmd)
            model_path = Path(find_spec(pkg.__model__).origin).parent
            return model_path
        except:
            msg.fail('Connection to server is failed.')
            msg.fail('Please try again later and make sure your Internet connection is on.')
            raise



import numpy as np
from spacy.tokens.doc import Doc
from spacy.tokens.span import Span
from spacy.tokens import Token as Token_

from .utils import lower, islower, isupper, istitle


class Token:
    """Class that encapsulates the spaCy Token object. Modifies and hides some token attributes.
    The spaCy Token object is available via `_token` attribute.  
    """

    def __init__(self, token: Token_):
        """
        Args:
            token (Token_): spaCy Token object.
            offsets: Tuple[int, str]: The character offset and trailing whitespaces 
                of the token within the document. 
        """
        self._token = token
        self.i = token.i

    def __len__(self):
        """The number of unicode characters in the token.
        """
        return self._token.__len__()

    def __unicode__(self):
        return self._token.__unicode__()

    def __bytes__(self):
        return self._token.__bytes__()

    def __str__(self):
        return self._token.__str__()

    def __repr__(self):
        return self._token.__repr__()

    @property
    def idx(self):
        """Returns:
            int: The character offset of the token within the document.
        """
        # set aligned token offset
        return self._token.doc._.idx[self.i]

    @property
    def pos(self):
        """
        Returns:
            str: Coarse-grained part-of-speech tag.
        """
        return self._token.tag_

    @property
    def lemma(self):
        """
        Returns:
            str: The token lemma.
        """
        return lower(self._token.lemma_)

    @property
    def is_sent_end(self):
        """
        Returns:
            bool: Whether the token ends a sentence.
        """
        return self._token._.sent_end

    @property
    def is_sent_start(self):
        """
        Returns:
            bool: Whether the token starts a sentence.
        """
        return bool(self._token.is_sent_start)

    @property
    def is_lower(self):
        """
        Returns:
            bool: Whether the token is in lowercase.
        """
        return islower(self._token.text)

    @property
    def is_upper(self):
        """
        Returns:
            bool: Whether the token is in uppercase.
        """
        return isupper(self._token.text)

    @property
    def is_title(self):
        """
        Returns:
            bool: Whether the token is in titlecase.
        """
        return istitle(self._token.text)

    @property
    def is_stop_word(self):
        """
        Returns:
            bool: Whether the token is a stop word.
        """
        return self._token.is_stop

    @property
    def vector(self):
        """Vector representation of the token. Token vector is the mean of the 
        subtoken vectors for transformer models. 

        Returns:
            numpy.ndarray[ndim=1, dtype='float32']: Token vector as 1D numpy array.
        """
        vector = self._token.vector
        if not vector.size and hasattr(self._token.doc._, 'trf_data'):
            trf_vector = []
            for i in self._token.doc._.trf_data.align[self.i].data:
                trf_vector.append(self._token.doc._.trf_data.tensors[0][0][i])
            vector = np.mean(np.array(trf_vector), axis=0)
        return vector

    @property
    def text(self):
        """
        Returns:
            str: The text of the token.
        """
        return self._token.text

    @property
    def text_with_ws(self):
        """
        Returns:
            str: The text of the token with trailing whitespaces if exist.
        """
        tws = self.text + self._token.doc._.ws[self.i]
        return tws if self.i > 0 else self._token.doc._.lws + tws

    @property
    def is_alpha(self):
        """
        Returns:
            bool: Whether the token consists of alpha characters.
        """
        return self._token.is_alpha

    @property
    def is_ascii(self):
        """
        Returns:
            bool: Whether the token consists of ASCII characters.
        """
        return self._token.is_ascii

    @property
    def is_bracket(self):
        """
        Returns:
            bool: Whether the token is a bracket.
        """
        return self._token.is_bracket

    @property
    def is_currency(self):
        """
        Returns:
            bool: Whether the token is a currency symbol.
        """
        return self._token.is_currency

    @property
    def is_digit(self):
        """
        Returns:
            bool: Whether the token consists of digits.
        """
        return self._token.is_digit

    @property
    def is_punct(self):
        """
        Returns:
            bool: Whether the token is punctuation.
        """
        return self._token.is_punct

    @property
    def is_quote(self):
        """
        Returns:
            bool: Whether the token is a quotation mark.
        """
        return self._token.is_quote

    @property
    def like_email(self):
        """
        Returns:
            bool: Whether the token resembles an email address.
        """
        return self._token.like_email

    @property
    def like_num(self):
        """
        Returns:
            bool: Whether the token resembles a number.
        """
        return self._token.like_num

    @property
    def like_url(self):
        """
        Returns:
            bool: Whether the token resembles a URL.
        """
        return self._token.like_url


class Sent:
    """Class that encapsulates the spaCy Span object. The spaCy Span object is available 
    via `_span` attribute.
    """

    def __init__(self, span: Span) -> None:
        """
        Args:
            span (Span): spaCy Span object.
        """
        self._span = span

    def __iter__(self):
        """Iterate over the tokens in the document.
        """
        for token in self._span:
            yield Token(token)

    def __getitem__(self, i: int) -> Token:
        """Return token at index `i`.

        Args:
            i (int): Index of the token.

        Returns:
            Token: Token object.
        """
        if not isinstance(i, int):
            raise ValueError('The attribute value should be an integer.')
        return Token(self._span.__getitem__(i))

    def __len__(self):
        """Return the number of tokens in the sentence.
        """
        return self._span.__len__()

    def __repr__(self):
        return self.text

    @property
    def text(self):
        """
        Returns:
            str: The text of the sentence.
        """
        return self.text_with_ws.rstrip()

    @property
    def text_with_ws(self):
        """
        Returns:
            str: The text of the sentence with trailing whitespaces if exist.
        """
        return ''.join(Token(t).text_with_ws for t in self._span)

    @property
    def start(self):
        """
        Returns:
            int: The index of the first token of the sentence. 
        """
        return self._span.start

    @property
    def end(self):
        """
        Returns:
            int: The index of the first token after the sentence. 
        """
        return self._span.end

    @property
    def vector(self):
        """Vector representation of the sentence as the mean of the token vectors.

        Returns:
            numpy.ndarray[ndim=1, dtype='float32']: Sentence vector as 1D numpy array.
        """
        vector = self._span.vector
        if not vector.size and hasattr(self._span.doc._, 'trf_data'):
            vector = sum(Token(t).vector for t in self._span) / len(self._span)
        return vector


class Document:
    """Class that encapsulates the spaCy Doc object. The spaCy Doc object is available 
    via `_doc` attribute.
    """

    def __init__(self, doc: Doc) -> None:
        """
        Args:
            doc (Doc): spaCy Doc object.
        """
        self._doc = doc

    def __iter__(self):
        """Iterate over the tokens in the document.
        """
        for token in self._doc:
            yield Token(token)

    def __getitem__(self, i: int) -> Token:
        """Return token at index `i`.

        Args:
            i (int): Index of the token.

        Returns:
            Token: Token object.
        """
        if not isinstance(i, int):
            raise ValueError('The attribute value should be an integer.')
        return Token(self._doc.__getitem__(i))

    def __len__(self):
        """Return the number of tokens in the document.
        """
        return self._doc.__len__()

    def __unicode__(self):
        return self.text

    def __bytes__(self):
        return self.text.encode('utf-8')

    def __str__(self):
        return self.__unicode__()

    def __repr__(self):
        return self.__str__()

    @property
    def sents(self):
        """Iterate over the sentences in the document.
        """
        if not self._doc.has_annotation('SENT_START'):
            yield Sent(Span(self._doc, 0, len(self._doc)))
        else:
            for sent in self._doc.sents:
                yield Sent(sent)

    @property
    def text(self):
        """
        Returns:
            str: The string representation of the document text. 
        """
        return ''.join(Token(t).text_with_ws for t in self._doc)

    @property
    def vector(self):
        """Vector representation of the document as the mean of the token vectors.

        Returns:
            numpy.ndarray[ndim=1, dtype='float32']: Sentence vector as 1D numpy array.
        """
        vector = self._doc.vector
        if not vector.size and hasattr(self._doc._, 'trf_data'):
            vector = sum(Token(t).vector for t in self._doc) / len(self._doc)
        return vector



import os
import glob
import shutil
import requests
from pathlib import Path
from zipfile import ZipFile
from typing import Iterator, Tuple, Iterable, Union, List, Any

from sklearn.model_selection import train_test_split
from wasabi import Printer

from .fs import FS


def lower(text: str) -> str:
    """Convert string to lowercase, fixes Turkish "İ" and "ı" problems.
    """
    return text.replace('I', 'ı').replace('İ', 'i').lower()


def upper(text: str) -> str:
    """Convert string to uppercase, fixes Turkish "İ" and "ı" problems.
    """
    return text.replace('i', 'İ').upper()


def capitalize(text: str) -> str:
    """Capitalize first character in string, fixes Turkish "İ" and "ı" problems.
    """
    text = lower(text)
    return text.replace(text[0], upper(text[0]), 1)


def capwords(text: str) -> str:
    """Capitalize each word in string, fixes Turkish "İ" and "ı" problems.
    """
    return ' '.join(capitalize(word) for word in text.split())


def islower(text: str) -> str:
    """Check if string contains all lowercase characters 
    """
    return text == lower(text)


def isupper(text: str) -> str:
    """Check if string contains all uppercase characters 
    """
    return text == upper(text)


def istitle(text: str) -> str:
    """Check if each word in string starts with an uppercase letter 
    """
    return all(word == capitalize(word) for word in text.split())


def batch_dataset(data: Iterable[Any], batch_size: int = 1) -> Iterator[Iterable[Any]]:
    """Split dataset into batches. Last batch would not be padded. 

    Args:
        data (Iterable[Any]): Data to be splitted.
        batch_size (int, optional): Batch size. Defaults to 1.

    Yields:
        Iterator[Iterable[Any]]: Yield batches, one for each batch.
    """
    if not isinstance(batch_size, int) or batch_size < 1:
        raise ValueError('"batch_size" must be a positive integer.')
    l = len(data)
    for ndx in range(0, l, batch_size):
        yield data[ndx:min(ndx + batch_size, l)]


def split_dataset(
    data: List[Any],
    split_ratios: Union[List[float], Tuple[float, ...]]
) -> Tuple[Union[List[Any], None], Union[List[Any], None], Union[List[Any], None]]:
    """Split dataset into train, dev, test sets.

    Args:
        data (List[Any]): Dataset to be splitted.
        split_ratios (Union[List[float], Tuple[float, ...]]): Train, dev, test split ratios 
            in the [0, 1] range. Either two or one of the ratios can be set to 0.0.

    Returns:
        Tuple[Union[List[Any], None], Union[List[Any], None], Union[List[Any], None]]: 
            Splitted train, dev, test sets.
    """
    if not data or not isinstance(data, list):
        raise ValueError('Data must be list of values.')

    if not isinstance(split_ratios, (tuple, list)) or len(split_ratios) != 3 or \
            not all([isinstance(v, float) and v >= 0. for v in split_ratios]) or \
            sum(split_ratios) != 1.0:
        raise ValueError(
            'Train, dev, test split ratios must be float in the [0, 1] range and the sum must be 1.0.')

    splits = {k: split_ratios[i] for i, k in enumerate(['train', 'dev', 'test'])}
    nr_splits = len([v for v in splits.values() if v])

    if nr_splits == 3:
        splits['dev'] = splits['dev']/(1 - splits['test'])
        train, test = train_test_split(data, test_size=splits['test'], random_state=42)
        train, dev = train_test_split(train, test_size=splits['dev'], random_state=42)
        return train, dev, test
    elif nr_splits == 2:
        if not splits['train']:
            dev, test = train_test_split(data, test_size=splits['test'],
                                         random_state=42)
            return None, dev, test
        elif not splits['dev']:
            train, test = train_test_split(data, test_size=splits['test'],
                                           random_state=42)
            return train, None, test
        else:
            train, dev = train_test_split(data, test_size=splits['dev'],
                                          random_state=42)
            return train, dev, None
    else:
        return (
            data if splits['train'] else None,
            data if splits['dev'] else None,
            data if splits['test'] else None
        )


def fetch_ud_treebanks(output_path: Union[str, Path]) -> None:
    """Fetch Universal Dependencies treebanks.

    Args:
        output_path (Union[str, Path]): Output path to save fetched UD treebanks.
    """
    base_url = 'https://github.com/UniversalDependencies/'
    repos = {
        'atis': 'UD_Turkish-Atis',
        'framenet': 'UD_Turkish-FrameNet',
        'kenet': 'UD_Turkish-Kenet',
        'penn': 'UD_Turkish-Penn'
    }

    for k, v in repos.items():
        url = base_url + v + '/archive/refs/heads/master.zip'

        repo_dir = os.path.join(output_path, k)
        os.makedirs(repo_dir, exist_ok=True)

        # create temporary directory to fetch all the files in repository
        tmp_dir = os.path.join(output_path, 'ud_tmp')
        os.makedirs(tmp_dir, exist_ok=True)
        zip_file = os.path.join(tmp_dir, 'ud.zip')

        try:
            res = requests.get(url, allow_redirects=True)
            with open(zip_file, 'wb') as f:
                f.write(res.content)
        except Exception as e:
            with open(os.path.join(output_path, 'error.log'), 'a+') as f:
                f.write(f'url: {url}\nmsg: {e}\n\n')

        with ZipFile(zip_file, 'r') as f:
            f.extractall(tmp_dir)

        for filepath in glob.glob(os.path.join(tmp_dir, '**', '*.conllu'), recursive=True):
            shutil.move(filepath, repo_dir)

        # remove temporary directory
        shutil.rmtree(tmp_dir, ignore_errors=True)


def merge_ud_treebanks(
    data_path: Union[str, Path],
    output_path: Union[str, Path],
    blacklist: Iterable[str] = None
) -> None:
    """Merge Universal Dependencies treebanks.

    Args:
        data_path (Union[str, Path]): Path to the UD treebanks.
        output_path (Union[str, Path]): Output path to save merged files. Merged dataset 
            would be splitted into train, dev, test sets if path is a directory.
        blacklist (Iterable[str], optional): Files and directories to be excluded. Defaults to None.
    """
    data = []
    for filepath in glob.glob(os.path.join(data_path, '**', '*.conllu'), recursive=True):
        if isinstance(blacklist, (list, tuple, set)) and \
                not all(n not in filepath.lstrip(str(data_path)) for n in blacklist):
            continue
        for lines in FS.read(filepath):
            ext, filename, _ = FS.split_path(filepath)
            # append source of each sentence as comment
            sent = [f'# source = {filename}.{ext}']
            for line in lines:
                line = line if isinstance(line, str) else '\t'.join(line)
                sent.append(line)
            data.append('\n'.join(sent))

    msg = Printer()
    msg.info(f'# of sentences: {len(data)}')

    if FS.split_path(output_path)[0]:
        # output is file
        FS.to_disk('\n\n'.join(data), output_path)
    else:
        # output is directory
        # split into train, dev, test sets
        train, dev, test = split_dataset(data, (0.8, 0.1, 0.1))
        FS.to_disk('\n\n'.join(train), os.path.join(output_path, 'train.conllu'))
        FS.to_disk('\n\n'.join(dev), os.path.join(output_path, 'dev.conllu'))
        FS.to_disk('\n\n'.join(test), os.path.join(output_path, 'test.conllu'))

        msg.info(f'# of train sentences: {len(train)}')
        msg.info(f'# of validation sentences: {len(dev)}')
        msg.info(f'# of test sentences: {len(test)}')



import re
from itertools import islice
from typing import Callable, Optional, Iterable, Tuple, Dict, List, Any

import numpy as np
from thinc.api import Model, SequenceCategoricalCrossentropy, Config
from spacy.tokens import Token
from spacy.tokens.doc import Doc
from spacy.pipeline.tagger import Tagger
from spacy.language import Language
from spacy.vocab import Vocab
from spacy.training.example import Example
from spacy.errors import Errors
from spacy.scorer import PRFScore
from spacy.training import validate_examples, validate_get_examples
from spacy.util import registry
from spacy import util


default_model_config = """
[model]
@architectures = "spacy.Tagger.v2"

[model.tok2vec]
@architectures = "spacy.HashEmbedCNN.v2"
pretrained_vectors = null
width = 96
depth = 4
embed_size = 2000
window_size = 1
maxout_pieces = 3
subword_features = true
"""
DEFAULT_SBD_MODEL = Config().from_str(default_model_config)["model"]

Token.set_extension("sent_end", default=False, force=True)
Token.set_extension("sbd_tag", default=False, force=True)
Token.set_extension("sbd_tag_", default=False, force=True)


@Language.factory(
    "sbd",
    assigns=["token._.sent_end", "token._.sbd_tag", "token._.sbd_tag_"],
    default_config={"model": DEFAULT_SBD_MODEL, "overwrite": False,
                    "scorer": {"@scorers": "spacy.sbd_scorer.v1"}, "neg_prefix": "!"},
    default_score_weights={"sbd_f": 1.0, "sbd_p": 0.0, "sbd_r": 0.0}
)
def make_sbd(
    nlp: Language,
    name: str,
    model: Model,
    overwrite: bool,
    scorer: Optional[Callable],
    neg_prefix: str,
):
    """Construct a sentence boundary detection component.
    """
    return SentenceBoundaryDetector(nlp.vocab, model, name, overwrite=overwrite,
                                    scorer=scorer, neg_prefix=neg_prefix)


def sbd_score(examples: Iterable[Example], **kwargs) -> Dict[str, float]:
    """Returns PRF scores for token tags.

    Args:
        examples (Iterable[Example]): Examples to score.

    Returns:
        Dict[str, float]: Dictionary containing the PRF scores.
    """
    micro_prf = PRFScore()
    for example in examples:
        gold_doc = example.reference
        pred_doc = example.predicted
        align = example.alignment
        gold_tags = set()
        for gold_i, token in enumerate(gold_doc):
            if token._.sbd_tag_ == 'EOS':
                gold_tags.add((gold_i, token._.sbd_tag))
        pred_tags = set()
        for token in pred_doc:
            if token.orth_.isspace():
                continue
            if align.x2y.lengths[token.i] == 1:
                gold_i = align.x2y[token.i][0]
                if token._.sbd_tag_ == 'EOS':
                    pred_tags.add((gold_i, token._.sbd_tag))
        micro_prf.score_set(pred_tags, gold_tags)
    return {"sbd_p": micro_prf.precision, "sbd_r": micro_prf.recall,
            "sbd_f": micro_prf.fscore}


@registry.scorers("spacy.sbd_scorer.v1")
def make_sbd_scorer():
    return sbd_score


class SentenceBoundaryDetector(Tagger):
    """Pipeline component for sentence boundary detection.
    """

    def __init__(
        self,
        vocab: Vocab,
        model: Model,
        name: str = "sbd",
        *,
        overwrite: bool = False,
        scorer: Optional[Callable] = sbd_score,
        neg_prefix: str = "!",
    ):
        """Initialize a sentence boundary detector.

        Args:
            vocab (Vocab): The shared vocabulary.
            model (Model): The Thinc Model powering the pipeline component.
            name (str): The component instance name, used to add entries to the
                losses during training.
            scorer (Optional[Callable]): The scoring method.
        """
        self.vocab = vocab
        self.model = model
        self.name = name
        self._rehearsal_model = None
        cfg = {"labels": [], "overwrite": overwrite, "neg_prefix": neg_prefix}
        self.cfg = dict(sorted(cfg.items()))
        self.scorer = scorer

    def get_aligned(self, example: Example) -> List[Any]:
        """
        Args:
            example (Example): Example object.

        Returns:
            List[Any]: Aligned array for a token attribute.
        """
        align = example.alignment.x2y
        vocab = example.reference.vocab
        gold_values = np.array([t._.sbd_tag for t in example.reference], dtype=np.uint64)
        output = [None] * len(example.predicted)
        for token in example.predicted:
            values = gold_values[align[token.i]]
            values = values.ravel()
            if len(values) == 0:
                output[token.i] = None
            elif len(values) == 1:
                output[token.i] = values[0]
            elif len(set(list(values))) == 1:
                # If all aligned tokens have the same value, use it.
                output[token.i] = values[0]
            else:
                output[token.i] = None
        return [vocab.strings[o] if o is not None else o for o in output]

    def set_annotations(self, docs: Iterable[Doc], batch_tag_ids):
        """Modify a batch of documents, using pre-computed scores.

        Args:
            docs (Iterable[Doc]): The documents to modify.
            batch_tag_ids: The IDs to set, produced by Tagger.predict.
        """
        if isinstance(docs, Doc):
            docs = [docs]
        labels = self.labels
        for i, doc in enumerate(docs):
            doc_tag_ids = batch_tag_ids[i]
            if hasattr(doc_tag_ids, "get"):
                doc_tag_ids = doc_tag_ids.get()
            eos_token = None
            for j, token in enumerate(doc):
                if eos_token:
                    if re.search(r'^[.?!:;)}\]]+$', token.text) or \
                            not re.search(r'[\w({\[]', token.text) and \
                            not re.search(r'^[.?!:;)}\]]+$', eos_token.text):
                        eos_token = token
                    else:
                        token.sent_start = True
                        eos_token._.sent_end = True
                        eos_token = None
                if token._.sbd_tag == 0 or self.cfg["overwrite"]:
                    token._.sbd_tag = self.vocab.strings[labels[doc_tag_ids[j]]]
                    token._.sbd_tag_ = labels[doc_tag_ids[j]]
                    if token._.sbd_tag_ == 'EOS':
                        eos_token = token
            token._.sent_end = True if eos_token else False

    def get_loss(self, examples: Iterable[Example], scores) -> Tuple[float, float]:
        """Find the loss and gradient of loss for the batch of documents and
        their predicted scores.

        Args:
            examples (Iterable[Example]): The batch of examples.
            scores: Scores representing the model's predictions.

        Returns:
            (Tuple[float, float]): The loss and the gradient.
        """
        validate_examples(examples, "SentenceBoundaryDetector.get_loss")
        loss_func = SequenceCategoricalCrossentropy(names=self.labels, normalize=False,
                                                    neg_prefix=self.cfg["neg_prefix"])
        # Convert empty tag "" to missing value None so that both misaligned tokens
        # and tokens with missing annotation have the default missing value None.
        truths = []
        for eg in examples:
            eg_truths = [tag if tag != "" else None
                         for tag in self.get_aligned(eg)]
            truths.append(eg_truths)
        d_scores, loss = loss_func(scores, truths)
        if self.model.ops.xp.isnan(loss):
            raise ValueError(Errors.E910.format(name=self.name))
        return float(loss), d_scores

    def initialize(
        self,
        get_examples: Callable[[], Iterable[Example]],
        *,
        nlp: Language = None,
        labels=None
    ) -> None:
        """Initialize the pipe for training, using a representative set
        of data examples.

        Args:
            get_examples (Callable[[], Iterable[Example]]): Function that
                returns a representative sample of gold-standard Example objects.
            nlp (Language): The current nlp object the component is part of.
            labels: The labels to add to the component, typically generated by the
                `init labels` command. If no labels are provided, the get_examples
                callback is used to extract the labels from the data.
        """
        validate_get_examples(
            get_examples, "SentenceBoundaryDetector.initialize")
        util.check_lexeme_norms(self.vocab, "sbd")
        if labels is not None:
            for tag in labels:
                self.add_label(tag)
        else:
            tags = set()
            for example in get_examples():
                for token in example.y:
                    if token._.sbd_tag_:
                        tags.add(token._.sbd_tag_)
            for tag in sorted(tags):
                self.add_label(tag)
        doc_sample = []
        label_sample = []
        for example in islice(get_examples(), 10):
            doc_sample.append(example.x)
            gold_tags = self.get_aligned(example)
            gold_array = [[1.0 if tag == gold_tag else 0.0 for tag in self.labels]
                          for gold_tag in gold_tags]
            label_sample.append(self.model.ops.asarray(
                gold_array, dtype="float32"))
        self._require_labels()
        assert len(doc_sample) > 0, Errors.E923.format(name=self.name)
        assert len(label_sample) > 0, Errors.E923.format(name=self.name)
        self.model.initialize(X=doc_sample, Y=label_sample)
